from ..base import *
try:
    from ..translators.translator import translate_js
except:
    pass


@Js
def Function():
    # convert arguments to python list of strings
    a = [e.to_string().value for e in arguments.to_list()]
    body = ';'
    args = ()
    if len(a):
        body = '%s;' % a[-1]
        args = a[:-1]
    # translate this function to js inline function
    js_func = '(function (%s) {%s})' % (','.join(args), body)
    # now translate js inline to python function
    py_func = translate_js(js_func, '')
    # add set func scope to global scope
    # a but messy solution but works :)
    globals()['var'] = PyJs.GlobalObject
    # define py function and return it
    temp = executor(py_func, globals())
    temp.source = '{%s}' % body
    temp.func_name = 'anonymous'
    return temp


def executor(f, glob):
    exec (f, globals())
    return globals()['PyJs_anonymous_0_']


#new statement simply calls Function
Function.create = Function

#set constructor property inside FunctionPrototype

fill_in_props(FunctionPrototype, {'constructor': Function}, default_attrs)

#attach prototype to Function constructor
Function.define_own_property(
    'prototype', {
        'value': FunctionPrototype,
        'enumerable': False,
        'writable': False,
        'configurable': False
    })
#Fix Function length (its 0 and should be 1)
Function.own['length']['value'] = Js(1)



from ..base import *

RegExpPrototype.define_own_property('constructor', {
    'value': RegExp,
    'enumerable': False,
    'writable': True,
    'configurable': True
})

RegExp.define_own_property(
    'prototype', {
        'value': RegExpPrototype,
        'enumerable': False,
        'writable': False,
        'configurable': False
    })



from .base import *
from .simplex import *


class Space(object):
    def __init__(self):
        self.GlobalObj = None
        self.ctx = None
        self.byte_generator = None

        self.Number = None
        self.String = None
        self.Boolean = None
        self.RegExp = None
        self.Object = None
        self.Array = None
        self.Function = None

        self.BooleanPrototype = None
        self.NumberPrototype = None
        self.StringPrototype = None

        self.FunctionPrototype = None
        self.ArrayPrototype = None
        self.RegExpPrototype = None
        self.DatePrototype = None
        self.ObjectPrototype = None

        self.ErrorPrototype = None
        self.EvalErrorPrototype = None
        self.RangeErrorPrototype = None
        self.ReferenceErrorPrototype = None
        self.SyntaxErrorPrototype = None
        self.TypeErrorPrototype = None
        self.URIErrorPrototype = None

        self.interpreter = None

    @property
    def ERROR_TYPES(self):
        return {
            'Error': self.ErrorPrototype,
            'EvalError': self.EvalErrorPrototype,
            'RangeError': self.RangeErrorPrototype,
            'ReferenceError': self.ReferenceErrorPrototype,
            'SyntaxError': self.SyntaxErrorPrototype,
            'TypeError': self.TypeErrorPrototype,
            'URIError': self.URIErrorPrototype,
        }

    def get_global_environment(self):
        return self.GlobalCtx.variable_environment()

    def NewObject(self):
        return PyJsObject(self.ObjectPrototype)

    def NewFunction(self, code, ctx, params, name, is_declaration,
                    definitions):
        return PyJsFunction(
            code,
            ctx,
            params,
            name,
            self,
            is_declaration,
            definitions,
            prototype=self.FunctionPrototype)

    def NewDate(self, value):
        return PyJsDate(value, self.DatePrototype)

    def NewArray(self, length=0):
        return PyJsArray(length, self.ArrayPrototype)

    def NewError(self, typ, message):
        return PyJsError(message, self.ERROR_TYPES[typ])

    def NewRegExp(self, body, flags):
        return PyJsRegExp(body, flags, self.RegExpPrototype)

    def ConstructArray(self, py_arr):
        ''' note py_arr elems are NOT converted to PyJs types!'''
        arr = self.NewArray(len(py_arr))
        arr._init(py_arr)
        return arr

    def ConstructObject(self, py_obj):
        ''' note py_obj items are NOT converted to PyJs types! '''
        obj = self.NewObject()
        for k, v in py_obj.items():
            obj.put(unicode(k), v)
        return obj



__all__ = ['require']

import subprocess, os, codecs, glob
from .evaljs import translate_js, DEFAULT_HEADER
from .translators.friendly_nodes import is_valid_py_name
import six
import tempfile
import hashlib
import random

DID_INIT = False
DIRNAME = tempfile.mkdtemp()
PY_NODE_MODULES_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'py_node_modules')


def _init():
    global DID_INIT
    if DID_INIT:
        return
    assert subprocess.call(
        'node -v', shell=True, cwd=DIRNAME
    ) == 0, 'You must have node installed! run: brew install node'
    assert subprocess.call(
        'cd %s;npm install babel-core babel-cli babel-preset-es2015 babel-polyfill babelify browserify browserify-shim'
        % repr(DIRNAME),
        shell=True,
        cwd=DIRNAME) == 0, 'Could not link required node_modules'
    DID_INIT = True


ADD_TO_GLOBALS_FUNC = '''
;function addToGlobals(name, obj) {
    if (!Object.prototype.hasOwnProperty('_fake_exports')) {
        Object.prototype._fake_exports = {};
    }
    Object.prototype._fake_exports[name] = obj;
};

'''
# subprocess.call("""node -e 'require("browserify")'""", shell=True)
GET_FROM_GLOBALS_FUNC = '''
;function getFromGlobals(name) {
    if (!Object.prototype.hasOwnProperty('_fake_exports')) {
        throw Error("Could not find any value named "+name);
    }
    if (Object.prototype._fake_exports.hasOwnProperty(name)) {
        return Object.prototype._fake_exports[name];
    } else {
        throw Error("Could not find any value named "+name);
    }
};

'''


def _get_module_py_name(module_name):
    return module_name.replace('-', '_')


def _get_module_var_name(module_name):
    cand =  _get_module_py_name(module_name).rpartition('/')[-1]
    if not is_valid_py_name(cand):
        raise ValueError(
            "Invalid Python module name %s (generated from %s). Unsupported/invalid npm module specification?" % (
                repr(cand), repr(module_name)))
    return cand


def _get_and_translate_npm_module(module_name, include_polyfill=False, update=False, maybe_version_str=""):
    assert isinstance(module_name, str), 'module_name must be a string!'

    py_name = _get_module_py_name(module_name)
    module_filename = '%s.py' % py_name
    var_name = _get_module_var_name(module_name)
    if not os.path.exists(os.path.join(PY_NODE_MODULES_PATH,
                                       module_filename)) or update:
        _init()
        module_hash = hashlib.sha1(module_name.encode("utf-8")).hexdigest()[:15]
        version = random.randrange(10000000000000)
        in_file_name = 'in_%s_%d.js' % (module_hash, version)
        out_file_name = 'out_%s_%d.js' % (module_hash, version)
        code = ADD_TO_GLOBALS_FUNC
        if include_polyfill:
            code += "\n;require('babel-polyfill');\n"
        code += """
        var module_temp_love_python = require(%s);
        addToGlobals(%s, module_temp_love_python);
        """ % (repr(module_name), repr(module_name))
        with open(os.path.join(DIRNAME, in_file_name), 'wb') as f:
            f.write(code.encode('utf-8') if six.PY3 else code)

        pkg_name = module_name.partition('/')[0]
        if maybe_version_str:
            pkg_name += '@' + maybe_version_str
        # make sure the module is installed
        assert subprocess.call(
            'cd %s;npm install %s' % (repr(DIRNAME), pkg_name),
            shell=True,
            cwd=DIRNAME
        ) == 0, 'Could not install the required module: ' + pkg_name

        # convert the module
        assert subprocess.call(
            '''node -e "(require('browserify')('./%s').bundle(function (err,data) {if (err) {console.log(err);throw new Error(err);};fs.writeFile('%s', require('babel-core').transform(data, {'presets': require('babel-preset-es2015')}).code, ()=>{});}))"'''
            % (in_file_name, out_file_name),
            shell=True,
            cwd=DIRNAME,
        ) == 0, 'Error when converting module to the js bundle'

        os.remove(os.path.join(DIRNAME, in_file_name))
        with codecs.open(os.path.join(DIRNAME, out_file_name), "r",
                         "utf-8") as f:
            js_code = f.read()
        print("Bundled JS library dumped at: %s" % os.path.join(DIRNAME, out_file_name))
        if len(js_code) < 50:
            raise RuntimeError("Candidate JS bundle too short - likely browserify issue.")
        js_code += GET_FROM_GLOBALS_FUNC
        js_code += ';var %s = getFromGlobals(%s);%s' % (
            var_name, repr(module_name), var_name)
        print('Please wait, translating...')
        py_code = translate_js(js_code)

        dirname = os.path.dirname(
            os.path.join(PY_NODE_MODULES_PATH, module_filename))
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        with open(os.path.join(PY_NODE_MODULES_PATH, module_filename),
                  'wb') as f:
            f.write(py_code.encode('utf-8') if six.PY3 else py_code)
    else:
        with codecs.open(
                os.path.join(PY_NODE_MODULES_PATH, module_filename), "r",
                "utf-8") as f:
            py_code = f.read()
    return py_code


def require(module_name, include_polyfill=True, update=False, context=None):
    """
    Installs the provided npm module, exports a js bundle via browserify, converts to ECMA 5.1 via babel and
    finally translates the generated JS bundle to Python via Js2Py.
    Returns a pure python object that behaves like the installed module. Nice!

    :param module_name: Name of the npm module to require. For example 'esprima'. Supports specific versions via @
        specification. Eg: 'crypto-js@3.3'.
    :param include_polyfill: Whether the babel-polyfill should be included as part of the translation. May be needed
    for some modules that use unsupported features of JS6 such as Map or typed arrays.
    :param update: Whether to force update the translation. Otherwise uses a cached version if exists.
    :param context: Optional context in which the translated module should be executed in. If provided, the
        header (js2py imports) will be skipped as it is assumed that the context already has all the necessary imports.
    :return: The JsObjectWrapper containing the translated module object. Can be used like a standard python object.
    """
    module_name, maybe_version = (module_name+"@@@").split('@')[:2]

    py_code = _get_and_translate_npm_module(module_name, include_polyfill=include_polyfill, update=update,
                                            maybe_version_str=maybe_version)
    # this is a bit hacky but we need to strip the default header from the generated code...
    if context is not None:
        if not py_code.startswith(DEFAULT_HEADER):
            # new header version? retranslate...
            assert not update, "Unexpected header."
            py_code = _get_and_translate_npm_module(module_name, include_polyfill=include_polyfill, update=True)
            assert py_code.startswith(DEFAULT_HEADER), "Unexpected header."
        py_code = py_code[len(DEFAULT_HEADER):]
    context = {} if context is None else context
    exec(py_code, context)
    return context['var'][_get_module_var_name(module_name)].to_py()



# todo make sure what they mean by desc undefined? None or empty? Answer: None :) it can never be empty but None is sometimes returned.

# I am implementing everything as dicts to speed up property creation

# Warning: value, get, set props of dest are PyJs types. Rest is Py!


def is_data_descriptor(desc):
    return desc and ('value' in desc or 'writable' in desc)


def is_accessor_descriptor(desc):
    return desc and ('get' in desc or 'set' in desc)


def is_generic_descriptor(
        desc
):  # generic means not the data and not the setter - therefore it must be one that changes only enum and conf
    return desc and not (is_data_descriptor(desc)
                         or is_accessor_descriptor(desc))


def from_property_descriptor(desc, space):
    if not desc:
        return {}
    obj = space.NewObject()
    if is_data_descriptor(desc):
        obj.define_own_property(
            'value', {
                'value': desc['value'],
                'writable': True,
                'enumerable': True,
                'configurable': True
            }, False)
        obj.define_own_property(
            'writable', {
                'value': desc['writable'],
                'writable': True,
                'enumerable': True,
                'configurable': True
            }, False)
    else:
        obj.define_own_property(
            'get', {
                'value': desc['get'],
                'writable': True,
                'enumerable': True,
                'configurable': True
            }, False)
        obj.define_own_property(
            'set', {
                'value': desc['set'],
                'writable': True,
                'enumerable': True,
                'configurable': True
            }, False)
    obj.define_own_property(
        'writable', {
            'value': desc['writable'],
            'writable': True,
            'enumerable': True,
            'configurable': True
        }, False)
    obj.define_own_property(
        'enumerable', {
            'value': desc['enumerable'],
            'writable': True,
            'enumerable': True,
            'configurable': True
        }, False)
    return obj


def to_property_descriptor(obj):
    if obj._type() != 'Object':
        raise TypeError()
    desc = {}
    for e in ('enumerable', 'configurable', 'writable'):
        if obj.has_property(e):
            desc[e] = obj.get(e).to_boolean().value
    if obj.has_property('value'):
        desc['value'] = obj.get('value')
    for e in ('get', 'set'):
        if obj.has_property(e):
            cand = obj.get(e)
            if not (cand.is_callable() or cand.is_undefined()):
                raise TypeError()
    if ('get' in desc or 'set' in desc) and ('value' in desc
                                             or 'writable' in desc):
        raise TypeError()



import binascii

from pyjsparser import PyJsParser
import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str

REGEXP_CONVERTER = PyJsParser()


def to_hex(s):
    return binascii.hexlify(s.encode('utf8')).decode(
        'utf8')  # fucking python 3, I hate it so much


    # wtf was wrong with s.encode('hex') ???
def indent(lines, ind=4):
    return ind * ' ' + lines.replace('\n', '\n' + ind * ' ').rstrip(' ')


def inject_before_lval(source, lval, code):
    if source.count(lval) > 1:
        print()
        print(lval)
        raise RuntimeError('To many lvals (%s)' % lval)
    elif not source.count(lval):
        print()
        print(lval)
        assert lval not in source
        raise RuntimeError('No lval found "%s"' % lval)
    end = source.index(lval)
    inj = source.rfind('\n', 0, end)
    ind = inj
    while source[ind + 1] == ' ':
        ind += 1
    ind -= inj
    return source[:inj + 1] + indent(code, ind) + source[inj + 1:]


def get_continue_label(label):
    return CONTINUE_LABEL % to_hex(label)


def get_break_label(label):
    return BREAK_LABEL % to_hex(label)


def is_valid_py_name(name):
    try:
        compile(name + ' =  11', 'a', 'exec')
    except:
        return False
    return True


def indent(lines, ind=4):
    return ind * ' ' + lines.replace('\n', '\n' + ind * ' ').rstrip(' ')


def compose_regex(val):
    reg, flags = val
    #reg = REGEXP_CONVERTER._unescape_string(reg)
    return u'/%s/%s' % (reg, flags)


def float_repr(f):
    if int(f) == f:
        return repr(int(f))
    return repr(f)


def argsplit(args, sep=','):
    """used to split JS args (it is not that simple as it seems because
       sep can be inside brackets).

       pass args *without* brackets!

       Used also to parse array and object elements, and more"""
    parsed_len = 0
    last = 0
    splits = []
    for e in bracket_split(args, brackets=['()', '[]', '{}']):
        if e[0] not in ('(', '[', '{'):
            for i, char in enumerate(e):
                if char == sep:
                    splits.append(args[last:parsed_len + i])
                    last = parsed_len + i + 1
        parsed_len += len(e)
    splits.append(args[last:])
    return splits


def bracket_split(source, brackets=('()', '{}', '[]'), strip=False):
    """DOES NOT RETURN EMPTY STRINGS (can only return empty bracket content if strip=True)"""
    starts = [e[0] for e in brackets]
    in_bracket = 0
    n = 0
    last = 0
    while n < len(source):
        e = source[n]
        if not in_bracket and e in starts:
            in_bracket = 1
            start = n
            b_start, b_end = brackets[starts.index(e)]
        elif in_bracket:
            if e == b_start:
                in_bracket += 1
            elif e == b_end:
                in_bracket -= 1
                if not in_bracket:
                    if source[last:start]:
                        yield source[last:start]
                    last = n + 1
                    yield source[start + strip:n + 1 - strip]
        n += 1
    if source[last:]:
        yield source[last:]


def js_comma(a, b):
    return 'PyJsComma(' + a + ',' + b + ')'


def js_or(a, b):
    return '(' + a + ' or ' + b + ')'


def js_bor(a, b):
    return '(' + a + '|' + b + ')'


def js_bxor(a, b):
    return '(' + a + '^' + b + ')'


def js_band(a, b):
    return '(' + a + '&' + b + ')'


def js_and(a, b):
    return '(' + a + ' and ' + b + ')'


def js_strict_eq(a, b):
    return 'PyJsStrictEq(' + a + ',' + b + ')'


def js_strict_neq(a, b):
    return 'PyJsStrictNeq(' + a + ',' + b + ')'


#Not handled by python in the same way like JS. For example 2==2==True returns false.
# In JS above would return true so we need brackets.
def js_abstract_eq(a, b):
    return '(' + a + '==' + b + ')'


#just like ==
def js_abstract_neq(a, b):
    return '(' + a + '!=' + b + ')'


def js_lt(a, b):
    return '(' + a + '<' + b + ')'


def js_le(a, b):
    return '(' + a + '<=' + b + ')'


def js_ge(a, b):
    return '(' + a + '>=' + b + ')'


def js_gt(a, b):
    return '(' + a + '>' + b + ')'


def js_in(a, b):
    return b + '.contains(' + a + ')'


def js_instanceof(a, b):
    return a + '.instanceof(' + b + ')'


def js_lshift(a, b):
    return '(' + a + '<<' + b + ')'


def js_rshift(a, b):
    return '(' + a + '>>' + b + ')'


def js_shit(a, b):
    return 'PyJsBshift(' + a + ',' + b + ')'


def js_add(
        a,
        b):  # To simplify later process of converting unary operators + and ++
    return '(%s+%s)' % (a, b)


def js_sub(a, b):  # To simplify
    return '(%s-%s)' % (a, b)


def js_mul(a, b):
    return '(' + a + '*' + b + ')'


def js_div(a, b):
    return '(' + a + '/' + b + ')'


def js_mod(a, b):
    return '(' + a + '%' + b + ')'


def js_typeof(a):
    cand = list(bracket_split(a, ('()', )))
    if len(cand) == 2 and cand[0] == 'var.get':
        return cand[0] + cand[1][:-1] + ',throw=False).typeof()'
    return a + '.typeof()'


def js_void(a):
    # eval and return undefined
    return 'PyJsComma(%s, Js(None))' % a


def js_new(a):
    cands = list(bracket_split(a, ('()', )))
    lim = len(cands)
    if lim < 2:
        return a + '.create()'
    n = 0
    while n < lim:
        c = cands[n]
        if c[0] == '(':
            if cands[n - 1].endswith(
                    '.get') and n + 1 >= lim:  # last get operation.
                return a + '.create()'
            elif cands[n - 1][0] == '(':
                return ''.join(cands[:n]) + '.create' + c + ''.join(
                    cands[n + 1:])
            elif cands[n - 1] == '.callprop':
                beg = ''.join(cands[:n - 1])
                args = argsplit(c[1:-1], ',')
                prop = args[0]
                new_args = ','.join(args[1:])
                create = '.get(%s).create(%s)' % (prop, new_args)
                return beg + create + ''.join(cands[n + 1:])
        n += 1
    return a + '.create()'


def js_delete(a):
    #replace last get with delete.
    c = list(bracket_split(a, ['()']))
    beg, arglist = ''.join(c[:-1]).strip(), c[-1].strip(
    )  #strips just to make sure... I will remove it later
    if beg[-4:] != '.get':
        print(a)
        raise SyntaxError('Invalid delete operation')
    return beg[:-3] + 'delete' + arglist


def js_neg(a):
    return '(-' + a + ')'


def js_pos(a):
    return '(+' + a + ')'


def js_inv(a):
    return '(~' + a + ')'


def js_not(a):
    return a + '.neg()'


def js_postfix(a, inc, post):
    bra = list(bracket_split(a, ('()', )))
    meth = bra[-2]
    if not meth.endswith('get'):
        raise SyntaxError('Invalid ++ or -- operation.')
    bra[-2] = bra[-2][:-3] + 'put'
    bra[-1] = '(%s,Js(%s.to_number())%sJs(1))' % (bra[-1][1:-1], a,
                                                  '+' if inc else '-')
    res = ''.join(bra)
    return res if not post else '(%s%sJs(1))' % (res, '-' if inc else '+')


def js_pre_inc(a):
    return js_postfix(a, True, False)


def js_post_inc(a):
    return js_postfix(a, True, True)


def js_pre_dec(a):
    return js_postfix(a, False, False)


def js_post_dec(a):
    return js_postfix(a, False, True)


CONTINUE_LABEL = 'JS_CONTINUE_LABEL_%s'
BREAK_LABEL = 'JS_BREAK_LABEL_%s'
PREPARE = '''HOLDER = var.own.get(NAME)\nvar.force_own_put(NAME, PyExceptionToJs(PyJsTempException))\n'''
RESTORE = '''if HOLDER is not None:\n    var.own[NAME] = HOLDER\nelse:\n    del var.own[NAME]\ndel HOLDER\n'''
TRY_CATCH = '''%stry:\nBLOCKfinally:\n%s''' % (PREPARE, indent(RESTORE))

OR = {'||': js_or}
AND = {'&&': js_and}
BOR = {'|': js_bor}
BXOR = {'^': js_bxor}
BAND = {'&': js_band}

EQS = {
    '===': js_strict_eq,
    '!==': js_strict_neq,
    '==': js_abstract_eq,  # we need == and != too. Read a note above method
    '!=': js_abstract_neq
}

#Since JS does not have chained comparisons we need to implement all cmp methods.
COMPS = {
    '<': js_lt,
    '<=': js_le,
    '>=': js_ge,
    '>': js_gt,
    'instanceof': js_instanceof,  #todo change to validitate
    'in': js_in
}

BSHIFTS = {'<<': js_lshift, '>>': js_rshift, '>>>': js_shit}

ADDS = {'+': js_add, '-': js_sub}

MULTS = {'*': js_mul, '/': js_div, '%': js_mod}
BINARY = {}
BINARY.update(ADDS)
BINARY.update(MULTS)
BINARY.update(BSHIFTS)
BINARY.update(COMPS)
BINARY.update(EQS)
BINARY.update(BAND)
BINARY.update(BXOR)
BINARY.update(BOR)
BINARY.update(AND)
BINARY.update(OR)
#Note they dont contain ++ and -- methods because they both have 2 different methods
# correct method will be found automatically in translate function
UNARY = {
    'typeof': js_typeof,
    'void': js_void,
    'new': js_new,
    'delete': js_delete,
    '!': js_not,
    '-': js_neg,
    '+': js_pos,
    '~': js_inv,
    '++': None,
    '--': None
}



print("import: 'js2py'")
import js2py

print("import: 'js2py.constructors'")
import js2py.constructors

print("import: 'js2py.es6'")
import js2py.es6

print("import: 'js2py.host'")
import js2py.host

print("import: 'js2py.internals'")
import js2py.internals

print("import: 'js2py.internals.constructors'")
import js2py.internals.constructors

print("import: 'js2py.internals.prototypes'")
import js2py.internals.prototypes

print("import: 'js2py.prototypes'")
import js2py.prototypes

print("import: 'js2py.py_node_modules'")
import js2py.py_node_modules

print("import: 'js2py.translators'")
import js2py.translators

print("import: 'js2py.utils'")
import js2py.utils




import six

if six.PY3:
    xrange = range
    import functools


def to_arr(this):
    """Returns Python array from Js array"""
    return [this.get(str(e)) for e in xrange(len(this))]


ARR_STACK = set({})


class ArrayPrototype:
    def toString():
        # this function is wrong but I will leave it here fore debugging purposes.
        func = this.get('join')
        if not func.is_callable():

            @this.Js
            def func():
                return '[object %s]' % this.Class

        return func.call(this, ())

    def toLocaleString():
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        # separator is simply a comma ','
        if not arr_len:
            return ''
        res = []
        for i in xrange(arr_len):
            element = array[str(i)]
            if element.is_undefined() or element.is_null():
                res.append('')
            else:
                cand = element.to_object()
                str_func = element.get('toLocaleString')
                if not str_func.is_callable():
                    raise this.MakeError(
                        'TypeError',
                        'toLocaleString method of item at index %d is not callable'
                        % i)
                res.append(element.callprop('toLocaleString').value)
        return ','.join(res)

    def concat():
        array = this.to_object()
        A = this.Js([])
        items = [array]
        items.extend(to_arr(arguments))
        n = 0
        for E in items:
            if E.Class == 'Array':
                k = 0
                e_len = len(E)
                while k < e_len:
                    if E.has_property(str(k)):
                        A.put(str(n), E.get(str(k)))
                    n += 1
                    k += 1
            else:
                A.put(str(n), E)
                n += 1
        return A

    def join(separator):
        ARR_STACK.add(this)
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        separator = ',' if separator.is_undefined() else separator.to_string(
        ).value
        elems = []
        for e in xrange(arr_len):
            elem = array.get(str(e))
            if elem in ARR_STACK:
                s = ''
            else:
                s = elem.to_string().value
            elems.append(
                s if not (elem.is_undefined() or elem.is_null()) else '')
        res = separator.join(elems)
        ARR_STACK.remove(this)
        return res

    def pop():  #todo check
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not arr_len:
            array.put('length', this.Js(arr_len))
            return None
        ind = str(arr_len - 1)
        element = array.get(ind)
        array.delete(ind)
        array.put('length', this.Js(arr_len - 1))
        return element

    def push(item):  # todo check
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        to_put = arguments.to_list()
        i = arr_len
        for i, e in enumerate(to_put, arr_len):
            array.put(str(i), e)
        if to_put:
            i += 1
            array.put('length', this.Js(i))
        return i

    def reverse():
        array = this.to_object()  # my own algorithm
        vals = to_arr(array)
        has_props = [array.has_property(str(e)) for e in xrange(len(array))]
        vals.reverse()
        has_props.reverse()
        for i, val in enumerate(vals):
            if has_props[i]:
                array.put(str(i), val)
            else:
                array.delete(str(i))
        return array

    def shift():  #todo check
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not arr_len:
            array.put('length', this.Js(0))
            return None
        first = array.get('0')
        for k in xrange(1, arr_len):
            from_s, to_s = str(k), str(k - 1)
            if array.has_property(from_s):
                array.put(to_s, array.get(from_s))
            else:
                array.delete(to)
        array.delete(str(arr_len - 1))
        array.put('length', this.Js(str(arr_len - 1)))
        return first

    def slice(start, end):  # todo check
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        relative_start = start.to_int()
        k = max((arr_len + relative_start), 0) if relative_start < 0 else min(
            relative_start, arr_len)
        relative_end = arr_len if end.is_undefined() else end.to_int()
        final = max((arr_len + relative_end), 0) if relative_end < 0 else min(
            relative_end, arr_len)
        res = []
        n = 0
        while k < final:
            pk = str(k)
            if array.has_property(pk):
                res.append(array.get(pk))
            k += 1
            n += 1
        return res

    def sort(cmpfn):
        if not this.Class in ('Array', 'Arguments'):
            return this.to_object()  # do nothing
        arr = []
        for i in xrange(len(this)):
            arr.append(this.get(six.text_type(i)))

        if not arr:
            return this
        if not cmpfn.is_callable():
            cmpfn = None
        cmp = lambda a, b: sort_compare(a, b, cmpfn)
        if six.PY3:
            key = functools.cmp_to_key(cmp)
            arr.sort(key=key)
        else:
            arr.sort(cmp=cmp)
        for i in xrange(len(arr)):
            this.put(six.text_type(i), arr[i])

        return this

    def splice(start, deleteCount):
        # 1-8
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        relative_start = start.to_int()
        actual_start = max(
            (arr_len + relative_start), 0) if relative_start < 0 else min(
                relative_start, arr_len)
        actual_delete_count = min(
            max(deleteCount.to_int(), 0), arr_len - actual_start)
        k = 0
        A = this.Js([])
        # 9
        while k < actual_delete_count:
            if array.has_property(str(actual_start + k)):
                A.put(str(k), array.get(str(actual_start + k)))
            k += 1
        # 10-11
        items = to_arr(arguments)[2:]
        items_len = len(items)
        # 12
        if items_len < actual_delete_count:
            k = actual_start
            while k < (arr_len - actual_delete_count):
                fr = str(k + actual_delete_count)
                to = str(k + items_len)
                if array.has_property(fr):
                    array.put(to, array.get(fr))
                else:
                    array.delete(to)
                k += 1
            k = arr_len
            while k > (arr_len - actual_delete_count + items_len):
                array.delete(str(k - 1))
                k -= 1
        # 13
        elif items_len > actual_delete_count:
            k = arr_len - actual_delete_count
            while k > actual_start:
                fr = str(k + actual_delete_count - 1)
                to = str(k + items_len - 1)
                if array.has_property(fr):
                    array.put(to, array.get(fr))
                else:
                    array.delete(to)
                k -= 1
        # 14-17
        k = actual_start
        while items:
            E = items.pop(0)
            array.put(str(k), E)
            k += 1
        array.put('length', this.Js(arr_len - actual_delete_count + items_len))
        return A

    def unshift():
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        argCount = len(arguments)
        k = arr_len
        while k > 0:
            fr = str(k - 1)
            to = str(k + argCount - 1)
            if array.has_property(fr):
                array.put(to, array.get(fr))
            else:
                array.delete(to)
            k -= 1
        j = 0
        items = to_arr(arguments)
        while items:
            E = items.pop(0)
            array.put(str(j), E)
            j += 1
        array.put('length', this.Js(arr_len + argCount))
        return arr_len + argCount

    def indexOf(searchElement):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if arr_len == 0:
            return -1
        if len(arguments) > 1:
            n = arguments[1].to_int()
        else:
            n = 0
        if n >= arr_len:
            return -1
        if n >= 0:
            k = n
        else:
            k = arr_len - abs(n)
            if k < 0:
                k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                elementK = array.get(str(k))
                if searchElement.strict_equality_comparison(elementK):
                    return k
            k += 1
        return -1

    def lastIndexOf(searchElement):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if arr_len == 0:
            return -1
        if len(arguments) > 1:
            n = arguments[1].to_int()
        else:
            n = arr_len - 1
        if n >= 0:
            k = min(n, arr_len - 1)
        else:
            k = arr_len - abs(n)
        while k >= 0:
            if array.has_property(str(k)):
                elementK = array.get(str(k))
                if searchElement.strict_equality_comparison(elementK):
                    return k
            k -= 1
        return -1

    def every(callbackfn):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                if not callbackfn.call(
                        T, (kValue, this.Js(k), array)).to_boolean().value:
                    return False
            k += 1
        return True

    def some(callbackfn):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                if callbackfn.call(
                        T, (kValue, this.Js(k), array)).to_boolean().value:
                    return True
            k += 1
        return False

    def forEach(callbackfn):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                callbackfn.call(T, (kValue, this.Js(k), array))
            k += 1

    def map(callbackfn):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        A = this.Js([])
        k = 0
        while k < arr_len:
            Pk = str(k)
            if array.has_property(Pk):
                kValue = array.get(Pk)
                mappedValue = callbackfn.call(T, (kValue, this.Js(k), array))
                A.define_own_property(
                    Pk, {
                        'value': mappedValue,
                        'writable': True,
                        'enumerable': True,
                        'configurable': True
                    })
            k += 1
        return A

    def filter(callbackfn):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        res = []
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                if callbackfn.call(
                        T, (kValue, this.Js(k), array)).to_boolean().value:
                    res.append(kValue)
            k += 1
        return res  # converted to js array automatically

    def reduce(callbackfn):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        if not arr_len and len(arguments) < 2:
            raise this.MakeError(
                'TypeError', 'Reduce of empty array with no initial value')
        k = 0
        if len(arguments) > 1:  # initial value present
            accumulator = arguments[1]
        else:
            kPresent = False
            while not kPresent and k < arr_len:
                kPresent = array.has_property(str(k))
                if kPresent:
                    accumulator = array.get(str(k))
                k += 1
            if not kPresent:
                raise this.MakeError(
                    'TypeError', 'Reduce of empty array with no initial value')
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                accumulator = callbackfn.call(
                    this.undefined, (accumulator, kValue, this.Js(k), array))
            k += 1
        return accumulator

    def reduceRight(callbackfn):
        array = this.to_object()
        arr_len = array.get('length').to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        if not arr_len and len(arguments) < 2:
            raise this.MakeError(
                'TypeError', 'Reduce of empty array with no initial value')
        k = arr_len - 1
        if len(arguments) > 1:  # initial value present
            accumulator = arguments[1]
        else:
            kPresent = False
            while not kPresent and k >= 0:
                kPresent = array.has_property(str(k))
                if kPresent:
                    accumulator = array.get(str(k))
                k -= 1
            if not kPresent:
                raise this.MakeError(
                    'TypeError', 'Reduce of empty array with no initial value')
        while k >= 0:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                accumulator = callbackfn.call(
                    this.undefined, (accumulator, kValue, this.Js(k), array))
            k -= 1
        return accumulator


def sort_compare(a, b, comp):
    if a is None:
        if b is None:
            return 0
        return 1
    if b is None:
        if a is None:
            return 0
        return -1
    if a.is_undefined():
        if b.is_undefined():
            return 0
        return 1
    if b.is_undefined():
        if a.is_undefined():
            return 0
        return -1
    if comp is not None:
        res = comp.call(a.undefined, (a, b))
        return res.to_int()
    x, y = a.to_string(), b.to_string()
    if x < y:
        return -1
    elif x > y:
        return 1
    return 0



from __future__ import unicode_literals
from ..conversions import *
from ..func_utils import *

import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str

RADIX_SYMBOLS = {
    0: '0',
    1: '1',
    2: '2',
    3: '3',
    4: '4',
    5: '5',
    6: '6',
    7: '7',
    8: '8',
    9: '9',
    10: 'a',
    11: 'b',
    12: 'c',
    13: 'd',
    14: 'e',
    15: 'f',
    16: 'g',
    17: 'h',
    18: 'i',
    19: 'j',
    20: 'k',
    21: 'l',
    22: 'm',
    23: 'n',
    24: 'o',
    25: 'p',
    26: 'q',
    27: 'r',
    28: 's',
    29: 't',
    30: 'u',
    31: 'v',
    32: 'w',
    33: 'x',
    34: 'y',
    35: 'z'
}


def to_str_rep(num):
    if is_nan(num):
        return 'NaN'
    elif is_infinity(num):
        sign = '-' if num < 0 else ''
        return sign + 'Infinity'
    elif int(num) == num:  # dont print .0
        return unicode(int(num))
    return unicode(num)  # todo: Make it 100% consistent with Node


class NumberPrototype:
    def toString(this, args):
        if GetClass(this) != 'Number':
            raise MakeError('TypeError',
                            'Number.prototype.valueOf is not generic')
        if type(this) != float:
            this = this.value
        radix = get_arg(args, 0)
        if is_undefined(radix):
            return to_str_rep(this)
        r = to_int(radix)
        if r == 10:
            return to_str_rep(this)
        if r not in xrange(2, 37) or radix != r:
            raise MakeError(
                'RangeError',
                'Number.prototype.toString() radix argument must be an integer between 2 and 36'
            )
        num = to_int(this)
        if num < 0:
            num = -num
            sign = '-'
        else:
            sign = ''
        res = ''
        while num:
            s = RADIX_SYMBOLS[num % r]
            num = num // r
            res = s + res
        return sign + (res if res else '0')

    def valueOf(this, args):
        if GetClass(this) != 'Number':
            raise MakeError('TypeError',
                            'Number.prototype.valueOf is not generic')
        if type(this) != float:
            this = this.value
        return this

    def toFixed(this, args):
        if GetClass(this) != 'Number':
            raise MakeError(
                'TypeError',
                'Number.prototype.toFixed called on incompatible receiver')
        if type(this) != float:
            this = this.value
        fractionDigits = get_arg(args, 0)
        digs = to_int(fractionDigits)
        if digs < 0 or digs > 20:
            raise MakeError(
                'RangeError',
                'toFixed() digits argument must be between 0 and 20')
        elif is_infinity(this):
            return 'Infinity' if this > 0 else '-Infinity'
        elif is_nan(this):
            return 'NaN'
        return format(this, '-.%df' % digs)

    def toExponential(this, args):
        if GetClass(this) != 'Number':
            raise MakeError(
                'TypeError',
                'Number.prototype.toExponential called on incompatible receiver'
            )
        if type(this) != float:
            this = this.value
        fractionDigits = get_arg(args, 0)
        digs = to_int(fractionDigits)
        if digs < 0 or digs > 20:
            raise MakeError(
                'RangeError',
                'toFixed() digits argument must be between 0 and 20')
        elif is_infinity(this):
            return 'Infinity' if this > 0 else '-Infinity'
        elif is_nan(this):
            return 'NaN'
        return format(this, '-.%de' % digs)

    def toPrecision(this, args):
        if GetClass(this) != 'Number':
            raise MakeError(
                'TypeError',
                'Number.prototype.toPrecision called on incompatible receiver')
        if type(this) != float:
            this = this.value
        precision = get_arg(args, 0)
        if is_undefined(precision):
            return to_string(this)
        prec = to_int(precision)
        if is_nan(this):
            return 'NaN'
        elif is_infinity(this):
            return 'Infinity' if this > 0 else '-Infinity'
        digs = prec - len(str(int(this)))
        if digs >= 0:
            return format(this, '-.%df' % digs)
        else:
            return format(this, '-.%df' % (prec - 1))


NumberPrototype.toLocaleString = NumberPrototype.toString



from __future__ import unicode_literals
import re

import datetime

from .desc import *
from .simplex import *
from .conversions import *

from pyjsparser import PyJsParser

import six
if six.PY2:
    from itertools import izip
else:
    izip = zip




def Type(obj):
    return obj.TYPE


# 8.6.2
class PyJs(object):
    TYPE = 'Object'
    IS_CONSTRUCTOR = False

    prototype = None
    Class = None
    extensible = True
    value = None

    own = {}

    def get_member(self, unconverted_prop):
        return self.get(to_string(unconverted_prop))

    def put_member(self, unconverted_prop, val):
        return self.put(to_string(unconverted_prop), val)

    def get(self, prop):
        assert type(prop) == unicode
        cand = self.get_property(prop)
        if cand is None:
            return undefined
        if is_data_descriptor(cand):
            return cand['value']
        if is_undefined(cand['get']):
            return undefined
        return cand['get'].call(self)

    def get_own_property(self, prop):
        assert type(prop) == unicode
        # takes py returns py
        return self.own.get(prop)

    def get_property(self, prop):
        assert type(prop) == unicode
        # take py returns py
        cand = self.get_own_property(prop)
        if cand:
            return cand
        if self.prototype is not None:
            return self.prototype.get_property(prop)

    def put(self, prop, val, throw=False):
        assert type(prop) == unicode
        # takes py, returns none
        if not self.can_put(prop):
            if throw:
                raise MakeError('TypeError', 'Could not define own property')
            return
        own_desc = self.get_own_property(prop)
        if is_data_descriptor(own_desc):
            self.own[prop]['value'] = val
            return
        desc = self.get_property(prop)
        if is_accessor_descriptor(desc):
            desc['set'].call(
                self, (val, ))  # calling setter on own or inherited element
        else:  # new property
            self.own[prop] = {
                'value': val,
                'writable': True,
                'configurable': True,
                'enumerable': True
            }

    def can_put(self, prop):  # to check
        assert type(prop) == unicode, type(prop)
        # takes py returns py
        desc = self.get_own_property(prop)
        if desc:  # if we have this property
            if is_accessor_descriptor(desc):
                return is_callable(
                    desc['set'])  # Check if setter method is defined
            else:  # data desc
                return desc['writable']
        if self.prototype is None:
            return self.extensible
        inherited = self.prototype.get_property(prop)
        if inherited is None:
            return self.extensible
        if is_accessor_descriptor(inherited):
            return not is_undefined(inherited['set'])
        elif self.extensible:
            return inherited['writable']  # weird...
        return False

    def has_property(self, prop):
        assert type(prop) == unicode
        # takes py returns Py
        return self.get_property(prop) is not None

    def delete(self, prop, throw=False):
        assert type(prop) == unicode
        # takes py, returns py
        desc = self.get_own_property(prop)
        if desc is None:
            return True
        if desc['configurable']:
            del self.own[prop]
            return True
        if throw:
            raise MakeError('TypeError', 'Could not define own property')
        return False

    def default_value(self, hint=None):
        order = ('valueOf', 'toString')
        if hint == 'String' or (hint is None and self.Class == 'Date'):
            order = ('toString', 'valueOf')
        for meth_name in order:
            method = self.get(meth_name)
            if method is not None and is_callable(method):
                cand = method.call(self, ())
                if is_primitive(cand):
                    return cand
        raise MakeError('TypeError',
                        'Cannot convert object to primitive value')

    def define_own_property(
            self, prop, desc,
            throw):  # Internal use only. External through Object
        assert type(prop) == unicode
        # takes Py, returns Py
        # prop must be a Py string. Desc is either a descriptor or accessor.
        # Messy method -  raw translation from Ecma spec to prevent any bugs. # todo check this
        current = self.get_own_property(prop)

        extensible = self.extensible
        if not current:  # We are creating a new OWN property
            if not extensible:
                if throw:
                    raise MakeError('TypeError',
                                    'Could not define own property')
                return False
            # extensible must be True
            if is_data_descriptor(desc) or is_generic_descriptor(desc):
                DEFAULT_DATA_DESC = {
                    'value': undefined,  # undefined
                    'writable': False,
                    'enumerable': False,
                    'configurable': False
                }
                DEFAULT_DATA_DESC.update(desc)
                self.own[prop] = DEFAULT_DATA_DESC
            else:
                DEFAULT_ACCESSOR_DESC = {
                    'get': undefined,  # undefined
                    'set': undefined,  # undefined
                    'enumerable': False,
                    'configurable': False
                }
                DEFAULT_ACCESSOR_DESC.update(desc)
                self.own[prop] = DEFAULT_ACCESSOR_DESC
            return True
        # therefore current exists!
        if not desc or desc == current:  # We don't need to change anything.
            return True
        configurable = current['configurable']
        if not configurable:  # Prevent changing params
            if desc.get('configurable'):
                if throw:
                    raise MakeError('TypeError',
                                    'Could not define own property')
                return False
            if 'enumerable' in desc and desc['enumerable'] != current[
                    'enumerable']:
                if throw:
                    raise MakeError('TypeError',
                                    'Could not define own property')
                return False
        if is_generic_descriptor(desc):
            pass
        elif is_data_descriptor(current) != is_data_descriptor(desc):
            # we want to change the current type of property
            if not configurable:
                if throw:
                    raise MakeError('TypeError',
                                    'Could not define own property')
                return False
            if is_data_descriptor(current):  # from data to setter
                del current['value']
                del current['writable']
                current['set'] = undefined  # undefined
                current['get'] = undefined  # undefined
            else:  # from setter to data
                del current['set']
                del current['get']
                current['value'] = undefined  # undefined
                current['writable'] = False
        elif is_data_descriptor(current) and is_data_descriptor(desc):
            if not configurable:
                if not current['writable'] and desc.get('writable'):
                    if throw:
                        raise MakeError('TypeError',
                                        'Could not define own property')
                    return False
            if not current['writable'] and 'value' in desc and current[
                    'value'] != desc['value']:
                if throw:
                    raise MakeError('TypeError',
                                    'Could not define own property')
                return False
        elif is_accessor_descriptor(current) and is_accessor_descriptor(desc):
            if not configurable:
                if 'set' in desc and desc['set'] != current['set']:
                    if throw:
                        raise MakeError('TypeError',
                                        'Could not define own property')
                    return False
                if 'get' in desc and desc['get'] != current['get']:
                    if throw:
                        raise MakeError('TypeError',
                                        'Could not define own property')
                    return False
        current.update(desc)
        return True

    def create(self, args, space):
        '''Generally not a constructor, raise an error'''
        raise MakeError('TypeError', '%s is not a constructor' % self.Class)


def get_member(
        self, prop, space
):  # general member getter, prop has to be unconverted prop. it is it can be any value
    typ = type(self)
    if typ not in PRIMITIVES:  # most likely getter for object
        return self.get_member(
            prop
        )  # <- object can implement this to support faster prop getting. ex array.
    elif typ == unicode:  # then probably a String
        if type(prop) == float and is_finite(prop):
            index = int(prop)
            if index == prop and 0 <= index < len(self):
                return self[index]
        s_prop = to_string(prop)
        if s_prop == 'length':
            return float(len(self))
        elif s_prop.isdigit():
            index = int(s_prop)
            if 0 <= index < len(self):
                return self[index]
        # use standard string prototype
        return space.StringPrototype.get(s_prop)
        # maybe an index
    elif typ == float:
        # use standard number prototype
        return space.NumberPrototype.get(to_string(prop))
    elif typ == bool:
        return space.BooleanPrototype.get(to_string(prop))
    elif typ is UNDEFINED_TYPE:
        raise MakeError('TypeError',
                        "Cannot read property '%s' of undefined" % prop)
    elif typ is NULL_TYPE:
        raise MakeError('TypeError',
                        "Cannot read property '%s' of null" % prop)
    else:
        raise RuntimeError('Unknown type! - ' + repr(typ))


def get_member_dot(self, prop, space):
    # dot member getter, prop has to be unicode
    typ = type(self)
    if typ not in PRIMITIVES:  # most likely getter for object
        return self.get(prop)
    elif typ == unicode:  # then probably a String
        if prop == 'length':
            return float(len(self))
        elif prop.isdigit():
            index = int(prop)
            if 0 <= index < len(self):
                return self[index]
        else:
            # use standard string prototype
            return space.StringPrototype.get(prop)
            # maybe an index
    elif typ == float:
        # use standard number prototype
        return space.NumberPrototype.get(prop)
    elif typ == bool:
        return space.BooleanPrototype.get(prop)
    elif typ in (UNDEFINED_TYPE, NULL_TYPE):
        raise MakeError('TypeError',
                        "Cannot read property '%s' of undefined" % prop)
    else:
        raise RuntimeError('Unknown type! - ' + repr(typ))


# Object


class PyJsObject(PyJs):
    TYPE = 'Object'
    Class = 'Object'

    def __init__(self, prototype=None):
        self.prototype = prototype
        self.own = {}

    def _init(self, props, vals):
        i = 0
        for prop, kind in props:
            if prop in self.own:  # just check... probably will not happen very often.
                if is_data_descriptor(self.own[prop]):
                    if kind != 'i':
                        raise MakeError(
                            'SyntaxError',
                            'Invalid object initializer! Duplicate property name "%s"'
                            % prop)
                else:
                    if kind == 'i' or (kind == 'g' and 'get' in self.own[prop]
                                       ) or (kind == 's'
                                             and 'set' in self.own[prop]):
                        raise MakeError(
                            'SyntaxError',
                            'Invalid object initializer! Duplicate setter/getter of prop: "%s"'
                            % prop)

            if kind == 'i':  # init
                self.own[prop] = {
                    'value': vals[i],
                    'writable': True,
                    'enumerable': True,
                    'configurable': True
                }
            elif kind == 'g':  # get
                self.define_own_property(prop, {
                    'get': vals[i],
                    'enumerable': True,
                    'configurable': True
                }, False)
            elif kind == 's':  # get
                self.define_own_property(prop, {
                    'get': vals[i],
                    'enumerable': True,
                    'configurable': True
                }, False)
            else:
                raise RuntimeError(
                    'Invalid property kind - %s. Expected one of i, g, s.' %
                    repr(kind))
            i += 1

    def _set_props(self, prop_descs):
        for prop, desc in six.iteritems(prop_descs):
            self.define_own_property(prop, desc)


# Array


# todo Optimise Array - extremely slow due to index conversions from str to int and back etc.
# solution - make get and put methods callable with any type of prop and handle conversions from inside
# if not array then use to_string(prop). In array if prop is integer then just use it
# also consider removal of these stupid writable, enumerable etc for ints.
class PyJsArray(PyJs):
    Class = 'Array'

    def __init__(self, length, prototype=None):
        self.prototype = prototype
        self.own = {
            'length': {
                'value': float(length),
                'writable': True,
                'enumerable': False,
                'configurable': False
            }
        }

    def _init(self, elements):
        for i, ele in enumerate(elements):
            if ele is None: continue
            self.own[unicode(i)] = {
                'value': ele,
                'writable': True,
                'enumerable': True,
                'configurable': True
            }

    def put(self, prop, val, throw=False):
        assert type(val) != int
        # takes py, returns none
        if not self.can_put(prop):
            if throw:
                raise MakeError('TypeError', 'Could not define own property')
            return
        own_desc = self.get_own_property(prop)
        if is_data_descriptor(own_desc):
            self.define_own_property(prop, {'value': val}, False)
            return
        desc = self.get_property(prop)
        if is_accessor_descriptor(desc):
            desc['set'].call(
                self, (val, ))  # calling setter on own or inherited element
        else:  # new property
            self.define_own_property(
                prop, {
                    'value': val,
                    'writable': True,
                    'configurable': True,
                    'enumerable': True
                }, False)

    def define_own_property(self, prop, desc, throw):
        assert type(desc.get('value')) != int
        old_len_desc = self.get_own_property('length')
        old_len = old_len_desc['value']  #  value is js type so convert to py.
        if prop == 'length':
            if 'value' not in desc:
                return PyJs.define_own_property(self, prop, desc, False)
            new_len = to_uint32(desc['value'])
            if new_len != to_number(desc['value']):
                raise MakeError('RangeError', 'Invalid range!')
            new_desc = dict((k, v) for k, v in six.iteritems(desc))
            new_desc['value'] = float(new_len)
            if new_len >= old_len:
                return PyJs.define_own_property(self, prop, new_desc, False)
            if not old_len_desc['writable']:
                return False
            if 'writable' not in new_desc or new_desc['writable'] == True:
                new_writable = True
            else:
                new_writable = False
                new_desc['writable'] = True
            if not PyJs.define_own_property(self, prop, new_desc, False):
                return False
            if new_len < old_len:
                # not very efficient for sparse arrays, so using different method for sparse:
                if old_len > 30 * len(self.own):
                    for ele in self.own.keys():
                        if ele.isdigit() and int(ele) >= new_len:
                            if not self.delete(
                                    ele
                            ):  # if failed to delete set len to current len and reject.
                                new_desc['value'] = old_len + 1.
                                if not new_writable:
                                    new_desc['writable'] = False
                                PyJs.define_own_property(
                                    self, prop, new_desc, False)
                                return False
                    old_len = new_len
                else:  # standard method
                    while new_len < old_len:
                        old_len -= 1
                        if not self.delete(
                                unicode(int(old_len))
                        ):  # if failed to delete set len to current len and reject.
                            new_desc['value'] = old_len + 1.
                            if not new_writable:
                                new_desc['writable'] = False
                            PyJs.define_own_property(self, prop, new_desc,
                                                     False)
                            return False
            if not new_writable:
                self.own['length']['writable'] = False
            return True

        elif prop.isdigit():
            index = to_uint32(prop)
            if index >= old_len and not old_len_desc['writable']:
                return False
            if not PyJs.define_own_property(self, prop, desc, False):
                return False
            if index >= old_len:
                old_len_desc['value'] = index + 1.
            return True
        else:
            return PyJs.define_own_property(self, prop, desc, False)

    def to_list(self):
        return [
            self.get(str(e)) for e in xrange(self.get('length').to_uint32())
        ]


# database with compiled patterns. Js pattern -> Py pattern.
REGEXP_DB = {}


class PyJsRegExp(PyJs):
    Class = 'RegExp'

    def __init__(self, body, flags, prototype=None):
        self.prototype = prototype
        self.glob = True if 'g' in flags else False
        self.ignore_case = re.IGNORECASE if 'i' in flags else 0
        self.multiline = re.MULTILINE if 'm' in flags else 0
        self.value = body

        if (body, flags) in REGEXP_DB:
            self.pat = REGEXP_DB[body, flags]
        else:
            comp = None
            try:
                # converting JS regexp pattern to Py pattern.
                possible_fixes = [(u'[]', u'[\0]'), (u'[^]', u'[^\0]'),
                                  (u'nofix1791', u'nofix1791')]
                reg = self.value
                for fix, rep in possible_fixes:
                    comp = PyJsParser()._interpret_regexp(reg, flags)
                    #print 'reg -> comp', reg, '->', comp
                    try:
                        self.pat = re.compile(
                            comp, self.ignore_case | self.multiline)
                        #print reg, '->', comp
                        break
                    except:
                        reg = reg.replace(fix, rep)
                    # print 'Fix', fix, '->', rep, '=', reg
                else:
                    raise Exception()
                REGEXP_DB[body, flags] = self.pat
            except:
                #print 'Invalid pattern...', self.value, comp
                raise MakeError(
                    'SyntaxError',
                    'Invalid RegExp pattern: %s -> %s' % (repr(self.value),
                                                          repr(comp)))
        # now set own properties:
        self.own = {
            'source': {
                'value': self.value,
                'enumerable': False,
                'writable': False,
                'configurable': False
            },
            'global': {
                'value': self.glob,
                'enumerable': False,
                'writable': False,
                'configurable': False
            },
            'ignoreCase': {
                'value': bool(self.ignore_case),
                'enumerable': False,
                'writable': False,
                'configurable': False
            },
            'multiline': {
                'value': bool(self.multiline),
                'enumerable': False,
                'writable': False,
                'configurable': False
            },
            'lastIndex': {
                'value': 0.,
                'enumerable': False,
                'writable': True,
                'configurable': False
            }
        }

    def match(self, string, pos):
        '''string is of course a py string'''
        return self.pat.match(string, int(pos))


class PyJsError(PyJs):
    Class = 'Error'
    extensible = True

    def __init__(self, message=None, prototype=None):
        self.prototype = prototype
        self.own = {}
        if message is not None:
            self.put('message', to_string(message))
            self.own['message']['enumerable'] = False


class PyJsDate(PyJs):
    Class = 'Date'
    UTCToLocal = None  # todo UTC to local should be imported!

    def __init__(self, value, prototype=None):
        self.value = value
        self.own = {}
        self.prototype = prototype

    # todo fix this problematic datetime part
    def to_local_dt(self):
        return datetime.datetime(1970, 1, 1) + datetime.timedelta(
            seconds=self.UTCToLocal(self.value) // 1000)

    def to_utc_dt(self):
        return datetime.datetime(1970, 1, 1) + datetime.timedelta(
            seconds=self.value // 1000)

    def local_strftime(self, pattern):
        if self.value is NaN:
            return 'Invalid Date'
        try:
            dt = self.to_local_dt()
        except:
            raise MakeError(
                'TypeError',
                'unsupported date range. Will fix in future versions')
        try:
            return dt.strftime(pattern)
        except:
            raise MakeError(
                'TypeError',
                'Could not generate date string from this date (limitations of python.datetime)'
            )

    def utc_strftime(self, pattern):
        if self.value is NaN:
            return 'Invalid Date'
        try:
            dt = self.to_utc_dt()
        except:
            raise MakeError(
                'TypeError',
                'unsupported date range. Will fix in future versions')
        try:
            return dt.strftime(pattern)
        except:
            raise MakeError(
                'TypeError',
                'Could not generate date string from this date (limitations of python.datetime)'
            )


# Scope class it will hold all the variables accessible to user
class Scope(PyJs):
    Class = 'Global'
    extensible = True
    IS_CHILD_SCOPE = True
    THIS_BINDING = None
    space = None
    exe = None

    # todo speed up!
    # in order to speed up this very important class the top scope should behave differently than
    # child scopes, child scope should not have this property descriptor thing because they cant be changed anyway
    # they are all confugurable= False

    def __init__(self, scope, space, parent=None):
        """Doc"""
        self.space = space
        self.prototype = parent
        if type(scope) is not dict:
            assert parent is not None, 'You initialised the WITH_SCOPE without a parent scope.'
            self.own = scope
            self.is_with_scope = True
        else:
            self.is_with_scope = False
        if parent is None:
            # global, top level scope
            self.own = {}
            for k, v in six.iteritems(scope):
                # set all the global items
                self.define_own_property(
                    k, {
                        'value': v,
                        'configurable': False,
                        'writable': False,
                        'enumerable': False
                    }, False)
        else:
            # not global, less powerful but faster closure.
            self.own = scope  # simple dictionary which maps name directly to js object.

        self.par = super(Scope, self)
        self.stack = []

    def register(self, var):
        # registered keeps only global registered variables
        if self.prototype is None:
            # define in global scope
            if var in self.own:
                self.own[var]['configurable'] = False
            else:
                self.define_own_property(
                    var, {
                        'value': undefined,
                        'configurable': False,
                        'writable': True,
                        'enumerable': True
                    }, False)
        elif var not in self.own:
            # define in local scope since it has not been defined yet
            self.own[var] = undefined  # default value

    def registers(self, vars):
        """register multiple variables"""
        for var in vars:
            self.register(var)

    def put(self, var, val, throw=False):
        if self.prototype is None:
            desc = self.own.get(var)  # global scope
            if desc is None:
                self.par.put(var, val, False)
            else:
                if desc['writable']:  # todo consider getters/setters
                    desc['value'] = val
        else:
            if self.is_with_scope:
                if self.own.has_property(var):
                    return self.own.put(var, val, throw=throw)
                else:
                    return self.prototype.put(var, val)
            # trying to put in local scope
            # we dont know yet in which scope we should place this var
            elif var in self.own:
                self.own[var] = val
                return val
            else:
                # try to put in the lower scope since we cant put in this one (var wasn't registered)
                return self.prototype.put(var, val)

    def get(self, var, throw=False):
        if self.prototype is not None:
            if self.is_with_scope:
                cand = None if not self.own.has_property(
                    var) else self.own.get(var)
            else:
                # fast local scope
                cand = self.own.get(var)
            if cand is None:
                return self.prototype.get(var, throw)
            return cand
        # slow, global scope
        if var not in self.own:
            # try in ObjectPrototype...
            if var in self.space.ObjectPrototype.own:
                return self.space.ObjectPrototype.get(var)
            if throw:
                raise MakeError('ReferenceError', '%s is not defined' % var)
            return undefined
        cand = self.own[var].get('value')
        return cand if cand is not None else self.own[var]['get'].call(self)

    def delete(self, var, throw=False):
        if self.prototype is not None:
            if self.is_with_scope:
                if self.own.has_property(var):
                    return self.own.delete(var)
            elif var in self.own:
                return False
            return self.prototype.delete(var)
        # we are in global scope here. Must exist and be configurable to delete
        if var not in self.own:
            # this var does not exist, why do you want to delete it???
            return True
        if self.own[var]['configurable']:
            del self.own[var]
            return True
        # not configurable, cant delete
        return False


def get_new_arguments_obj(args, space):
    obj = space.NewObject()
    obj.Class = 'Arguments'
    obj.define_own_property(
        'length', {
            'value': float(len(args)),
            'writable': True,
            'enumerable': False,
            'configurable': True
        }, False)
    for i, e in enumerate(args):
        obj.put(unicode(i), e)
    return obj


#Function
class PyJsFunction(PyJs):
    Class = 'Function'
    source = '{ [native code] }'
    IS_CONSTRUCTOR = True

    def __init__(self,
                 code,
                 ctx,
                 params,
                 name,
                 space,
                 is_declaration,
                 definitions,
                 prototype=None):
        self.prototype = prototype
        self.own = {}

        self.code = code
        if type(
                self.code
        ) == int:  # just a label pointing to the beginning of the code.
            self.is_native = False
        else:
            self.is_native = True  # python function

        self.ctx = ctx

        self.params = params
        self.arguments_in_params = 'arguments' in params
        self.definitions = definitions

        # todo remove this check later
        for p in params:
            assert p in self.definitions

        self.name = name
        self.space = space
        self.is_declaration = is_declaration

        #set own property length to the number of arguments
        self.own['length'] = {
            'value': float(len(params)),
            'writable': False,
            'enumerable': False,
            'configurable': False
        }

        if name:
            self.own['name'] = {
                'value': name,
                'writable': False,
                'enumerable': False,
                'configurable': True
            }

        if not self.is_native:  # set prototype for user defined functions
            # constructor points to this function
            proto = space.NewObject()
            proto.own['constructor'] = {
                'value': self,
                'writable': True,
                'enumerable': False,
                'configurable': True
            }
            self.own['prototype'] = {
                'value': proto,
                'writable': True,
                'enumerable': False,
                'configurable': False
            }
        # todo set up throwers on callee and arguments if in strict mode

    def call(self, this, args=()):
        ''' Dont use this method from inside bytecode to call other bytecode. '''
        if self.is_native:
            _args = SpaceTuple(
                args
            )  # we have to do that unfortunately to pass all the necessary info to the funcs
            _args.space = self.space
            return self.code(
                this, _args
            )  # must return valid js object - undefined, null, float, unicode, bool, or PyJs
        else:
            return self.space.exe._call(self, this,
                                        args)  # will run inside bytecode

    def has_instance(self, other):
        # I am not sure here so instanceof may not work lol.
        if not is_object(other):
            return False
        proto = self.get('prototype')
        if not is_object(proto):
            raise MakeError(
                'TypeError',
                'Function has non-object prototype in instanceof check')
        while True:
            other = other.prototype
            if not other:  # todo make sure that the condition is not None or null
                return False
            if other is proto:
                return True

    def create(self, args, space):
        proto = self.get('prototype')
        if not is_object(proto):
            proto = space.ObjectPrototype
        new = PyJsObject(prototype=proto)
        res = self.call(new, args)
        if is_object(res):
            return res
        return new

    def _generate_my_context(self, this, args):
        my_ctx = Scope(
            dict(izip(self.params, args)), self.space, parent=self.ctx)
        my_ctx.registers(self.definitions)
        my_ctx.THIS_BINDING = this
        if not self.arguments_in_params:
            my_ctx.own['arguments'] = get_new_arguments_obj(args, self.space)
        if not self.is_declaration and self.name and self.name not in my_ctx.own:
            my_ctx.own[
                self.
                name] = self  # this should be immutable binding but come on!
        return my_ctx


class SpaceTuple:
    def __init__(self, tup):
        self.tup = tup

    def __len__(self):
        return len(self.tup)

    def __getitem__(self, item):
        return self.tup[item]

    def __iter__(self):
        return iter(self.tup)



from __future__ import unicode_literals
from ..conversions import *
from ..func_utils import *
from ..operations import strict_equality_op
import json

indent = ''
# python 3 support
import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str


def parse(this, args):
    text, reviver = get_arg(args, 0), get_arg(args, 1)
    s = to_string(text)
    try:
        unfiltered = json.loads(s)
    except:
        raise MakeError(
            'SyntaxError',
            'JSON.parse could not parse JSON string - Invalid syntax')
    unfiltered = to_js(unfiltered, args.space)
    if is_callable(reviver):
        root = args.space.ConstructObject({'': unfiltered})
        return walk(root, '', reviver)
    else:
        return unfiltered


def stringify(this, args):
    global indent
    value, replacer, space = get_arg(args, 0), get_arg(args, 1), get_arg(
        args, 2)
    stack = set([])
    indent = ''
    property_list = replacer_function = undefined
    if is_object(replacer):
        if is_callable(replacer):
            replacer_function = replacer
        elif replacer.Class == 'Array':
            property_list = []
            for e in replacer:
                v = replacer[e]
                item = undefined
                typ = Type(v)
                if typ == 'Number':
                    item = to_string(v)
                elif typ == 'String':
                    item = v
                elif typ == 'Object':
                    if GetClass(v) in ('String', 'Number'):
                        item = to_string(v)
                if not is_undefined(item) and item not in property_list:
                    property_list.append(item)
    if is_object(space):
        if GetClass(space) == 'Number':
            space = to_number(space)
        elif GetClass(space) == 'String':
            space = to_string(space)
    if Type(space) == 'Number':
        space = min(10, to_int(space))
        gap = max(int(space), 0) * ' '
    elif Type(space) == 'String':
        gap = space[:10]
    else:
        gap = ''
    return Str('', args.space.ConstructObject({
        '': value
    }), replacer_function, property_list, gap, stack, space)


def Str(key, holder, replacer_function, property_list, gap, stack, space):
    value = holder.get(key)
    if is_object(value):
        to_json = value.get('toJSON')
        if is_callable(to_json):
            value = to_json.call(value, (key, ))
    if not is_undefined(replacer_function):
        value = replacer_function.call(holder, (key, value))

    if is_object(value):
        if value.Class == 'String':
            value = to_string(value)
        elif value.Class == 'Number':
            value = to_number(value)
        elif value.Class == 'Boolean':
            value = to_boolean(value)
    typ = Type(value)
    if is_null(value):
        return 'null'
    elif typ == 'Boolean':
        return 'true' if value else 'false'
    elif typ == 'String':
        return Quote(value)
    elif typ == 'Number':
        if not is_infinity(value):
            return to_string(value)
        return 'null'
    if is_object(value) and not is_callable(value):
        if value.Class == 'Array':
            return ja(value, stack, gap, property_list, replacer_function,
                      space)
        else:
            return jo(value, stack, gap, property_list, replacer_function,
                      space)
    return undefined


def jo(value, stack, gap, property_list, replacer_function, space):
    global indent
    if value in stack:
        raise MakeError('TypeError', 'Converting circular structure to JSON')
    stack.add(value)
    stepback = indent
    indent += gap
    if not is_undefined(property_list):
        k = property_list
    else:
        k = [unicode(e) for e, d in value.own.items() if d.get('enumerable')]
    partial = []
    for p in k:
        str_p = Str(p, value, replacer_function, property_list, gap, stack,
                    space)
        if not is_undefined(str_p):
            member = json.dumps(p) + ':' + (
                ' ' if gap else
                '') + str_p  # todo not sure here - what space character?
            partial.append(member)
    if not partial:
        final = '{}'
    else:
        if not gap:
            final = '{%s}' % ','.join(partial)
        else:
            sep = ',\n' + indent
            properties = sep.join(partial)
            final = '{\n' + indent + properties + '\n' + stepback + '}'
    stack.remove(value)
    indent = stepback
    return final


def ja(value, stack, gap, property_list, replacer_function, space):
    global indent
    if value in stack:
        raise MakeError('TypeError', 'Converting circular structure to JSON')
    stack.add(value)
    stepback = indent
    indent += gap
    partial = []
    length = js_arr_length(value)
    for index in xrange(length):
        index = unicode(index)
        str_index = Str(index, value, replacer_function, property_list, gap,
                        stack, space)
        if is_undefined(str_index):
            partial.append('null')
        else:
            partial.append(str_index)
    if not partial:
        final = '[]'
    else:
        if not gap:
            final = '[%s]' % ','.join(partial)
        else:
            sep = ',\n' + indent
            properties = sep.join(partial)
            final = '[\n' + indent + properties + '\n' + stepback + ']'
    stack.remove(value)
    indent = stepback
    return final


def Quote(string):
    return json.dumps(string)


def to_js(d, _args_space):
    return convert_to_js_type(d, _args_space)


def walk(holder, name, reviver):
    val = holder.get(name)
    if GetClass(val) == 'Array':
        for i in xrange(js_arr_length(val)):
            i = unicode(i)
            new_element = walk(val, i, reviver)
            if is_undefined(new_element):
                val.delete(i)
            else:
                new_element.put(i, new_element)
    elif is_object(val):
        for key in [
                unicode(e) for e, d in val.own.items() if d.get('enumerable')
        ]:
            new_element = walk(val, key, reviver)
            if is_undefined(new_element):
                val.delete(key)
            else:
                val.put(key, new_element)
    return reviver.call(holder, (name, val))



from __future__ import unicode_literals
from ..conversions import *
from ..func_utils import *
from ..operations import strict_equality_op

import six

if six.PY3:
    xrange = range
    import functools

ARR_STACK = set({})


class ArrayPrototype:
    def toString(this, args):
        arr = to_object(this, args.space)
        func = arr.get('join')
        if not is_callable(func):
            return u'[object %s]' % GetClass(arr)
        return func.call(this, ())

    def toLocaleString(this, args):
        array = to_object(this, args.space)
        arr_len = js_arr_length(array)
        # separator is simply a comma ','
        if not arr_len:
            return ''
        res = []
        for i in xrange(arr_len):
            element = array.get(unicode(i))
            if is_undefined(element) or is_null(element):
                res.append('')
            else:
                cand = to_object(element, args.space)
                str_func = cand.get('toLocaleString')
                if not is_callable(str_func):
                    raise MakeError(
                        'TypeError',
                        'toLocaleString method of item at index %d is not callable'
                        % i)
                res.append(to_string(str_func.call(cand, ())))
        return ','.join(res)

    def concat(this, args):
        array = to_object(this, args.space)
        items = [array]
        items.extend(tuple(args))
        A = []
        for E in items:
            if GetClass(E) == 'Array':
                k = 0
                e_len = js_arr_length(E)
                while k < e_len:
                    if E.has_property(unicode(k)):
                        A.append(E.get(unicode(k)))
                    k += 1
            else:
                A.append(E)
        return args.space.ConstructArray(A)

    def join(this, args):
        ARR_STACK.add(this)
        array = to_object(this, args.space)
        separator = get_arg(args, 0)
        arr_len = js_arr_length(array)
        separator = ',' if is_undefined(separator) else to_string(separator)
        elems = []
        for e in xrange(arr_len):
            elem = array.get(unicode(e))
            if elem in ARR_STACK:
                s = ''
            else:
                s = to_string(elem)
            elems.append(
                s if not (is_undefined(elem) or is_null(elem)) else '')
        res = separator.join(elems)
        ARR_STACK.remove(this)
        return res

    def pop(this, args):  #todo check
        array = to_object(this, args.space)
        arr_len = js_arr_length(array)
        if not arr_len:
            array.put('length', float(arr_len))
            return undefined
        ind = unicode(arr_len - 1)
        element = array.get(ind)
        array.delete(ind)
        array.put('length', float(arr_len - 1))
        return element

    def push(this, args):
        array = to_object(this, args.space)
        arr_len = js_arr_length(array)
        to_put = tuple(args)
        i = arr_len
        for i, e in enumerate(to_put, arr_len):
            array.put(unicode(i), e, True)
        array.put('length', float(arr_len + len(to_put)), True)
        return float(i)

    def reverse(this, args):
        array = to_object(this, args.space)
        vals = js_array_to_list(array)
        has_props = [
            array.has_property(unicode(e))
            for e in xrange(js_arr_length(array))
        ]
        vals.reverse()
        has_props.reverse()
        for i, val in enumerate(vals):
            if has_props[i]:
                array.put(unicode(i), val)
            else:
                array.delete(unicode(i))
        return array

    def shift(this, args):
        array = to_object(this, args.space)
        arr_len = js_arr_length(array)
        if not arr_len:
            array.put('length', 0.)
            return undefined
        first = array.get('0')
        for k in xrange(1, arr_len):
            from_s, to_s = unicode(k), unicode(k - 1)
            if array.has_property(from_s):
                array.put(to_s, array.get(from_s))
            else:
                array.delete(to_s)
        array.delete(unicode(arr_len - 1))
        array.put('length', float(arr_len - 1))
        return first

    def slice(this, args):  # todo check
        array = to_object(this, args.space)
        start = get_arg(args, 0)
        end = get_arg(args, 1)
        arr_len = js_arr_length(array)
        relative_start = to_int(start)
        k = max((arr_len + relative_start), 0) if relative_start < 0 else min(
            relative_start, arr_len)
        relative_end = arr_len if is_undefined(end) else to_int(end)
        final = max((arr_len + relative_end), 0) if relative_end < 0 else min(
            relative_end, arr_len)
        res = []
        n = 0
        while k < final:
            pk = unicode(k)
            if array.has_property(pk):
                res.append(array.get(pk))
            k += 1
            n += 1
        return args.space.ConstructArray(res)

    def sort(
            this, args
    ):  # todo: this assumes array continous (not sparse) - fix for sparse arrays
        cmpfn = get_arg(args, 0)
        if not GetClass(this) in ('Array', 'Arguments'):
            return to_object(this, args.space)  # do nothing
        arr_len = js_arr_length(this)
        if not arr_len:
            return this
        arr = [
            (this.get(unicode(e)) if this.has_property(unicode(e)) else None)
            for e in xrange(arr_len)
        ]
        if not is_callable(cmpfn):
            cmpfn = None
        cmp = lambda a, b: sort_compare(a, b, cmpfn)
        if six.PY3:
            key = functools.cmp_to_key(cmp)
            arr.sort(key=key)
        else:
            arr.sort(cmp=cmp)
        for i in xrange(arr_len):
            if arr[i] is None:
                this.delete(unicode(i))
            else:
                this.put(unicode(i), arr[i])
        return this

    def splice(this, args):
        # 1-8
        array = to_object(this, args.space)
        start = get_arg(args, 0)
        deleteCount = get_arg(args, 1)
        arr_len = js_arr_length(this)
        relative_start = to_int(start)
        actual_start = max(
            (arr_len + relative_start), 0) if relative_start < 0 else min(
                relative_start, arr_len)
        actual_delete_count = min(
            max(to_int(deleteCount), 0), arr_len - actual_start)
        k = 0
        A = args.space.NewArray(0)
        # 9
        while k < actual_delete_count:
            if array.has_property(unicode(actual_start + k)):
                A.put(unicode(k), array.get(unicode(actual_start + k)))
            k += 1
        # 10-11
        items = list(args)[2:]
        items_len = len(items)
        # 12
        if items_len < actual_delete_count:
            k = actual_start
            while k < (arr_len - actual_delete_count):
                fr = unicode(k + actual_delete_count)
                to = unicode(k + items_len)
                if array.has_property(fr):
                    array.put(to, array.get(fr))
                else:
                    array.delete(to)
                k += 1
            k = arr_len
            while k > (arr_len - actual_delete_count + items_len):
                array.delete(unicode(k - 1))
                k -= 1
        # 13
        elif items_len > actual_delete_count:
            k = arr_len - actual_delete_count
            while k > actual_start:
                fr = unicode(k + actual_delete_count - 1)
                to = unicode(k + items_len - 1)
                if array.has_property(fr):
                    array.put(to, array.get(fr))
                else:
                    array.delete(to)
                k -= 1
        # 14-17
        k = actual_start
        while items:
            E = items.pop(0)
            array.put(unicode(k), E)
            k += 1
        array.put('length', float(arr_len - actual_delete_count + items_len))
        return A

    def unshift(this, args):
        array = to_object(this, args.space)
        arr_len = js_arr_length(array)
        argCount = len(args)
        k = arr_len
        while k > 0:
            fr = unicode(k - 1)
            to = unicode(k + argCount - 1)
            if array.has_property(fr):
                array.put(to, array.get(fr))
            else:
                array.delete(to)
            k -= 1
        items = tuple(args)
        for j, e in enumerate(items):
            array.put(unicode(j), e)
        array.put('length', float(arr_len + argCount))
        return float(arr_len + argCount)

    def indexOf(this, args):
        array = to_object(this, args.space)
        searchElement = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if arr_len == 0:
            return -1.
        if len(args) > 1:
            n = to_int(args[1])
        else:
            n = 0
        if n >= arr_len:
            return -1.
        if n >= 0:
            k = n
        else:
            k = arr_len - abs(n)
            if k < 0:
                k = 0
        while k < arr_len:
            if array.has_property(unicode(k)):
                elementK = array.get(unicode(k))
                if strict_equality_op(searchElement, elementK):
                    return float(k)
            k += 1
        return -1.

    def lastIndexOf(this, args):
        array = to_object(this, args.space)
        searchElement = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if arr_len == 0:
            return -1.
        if len(args) > 1:
            n = to_int(args[1])
        else:
            n = arr_len - 1
        if n >= 0:
            k = min(n, arr_len - 1)
        else:
            k = arr_len - abs(n)
        while k >= 0:
            if array.has_property(unicode(k)):
                elementK = array.get(unicode(k))
                if strict_equality_op(searchElement, elementK):
                    return float(k)
            k -= 1
        return -1.

    def every(this, args):
        array = to_object(this, args.space)
        callbackfn = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if not is_callable(callbackfn):
            raise MakeError('TypeError', 'callbackfn must be a function')
        T = get_arg(args, 1)
        k = 0
        while k < arr_len:
            if array.has_property(unicode(k)):
                kValue = array.get(unicode(k))
                if not to_boolean(
                        callbackfn.call(T, (kValue, float(k), array))):
                    return False
            k += 1
        return True

    def some(this, args):
        array = to_object(this, args.space)
        callbackfn = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if not is_callable(callbackfn):
            raise MakeError('TypeError', 'callbackfn must be a function')
        T = get_arg(args, 1)
        k = 0
        while k < arr_len:
            if array.has_property(unicode(k)):
                kValue = array.get(unicode(k))
                if to_boolean(callbackfn.call(T, (kValue, float(k), array))):
                    return True
            k += 1
        return False

    def forEach(this, args):
        array = to_object(this, args.space)
        callbackfn = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if not is_callable(callbackfn):
            raise MakeError('TypeError', 'callbackfn must be a function')
        _this = get_arg(args, 1)
        k = 0
        while k < arr_len:
            sk = unicode(k)
            if array.has_property(sk):
                kValue = array.get(sk)
                callbackfn.call(_this, (kValue, float(k), array))
            k += 1
        return undefined

    def map(this, args):
        array = to_object(this, args.space)
        callbackfn = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if not is_callable(callbackfn):
            raise MakeError('TypeError', 'callbackfn must be a function')
        _this = get_arg(args, 1)
        k = 0
        A = args.space.NewArray(0)
        while k < arr_len:
            Pk = unicode(k)
            if array.has_property(Pk):
                kValue = array.get(Pk)
                mappedValue = callbackfn.call(_this, (kValue, float(k), array))
                A.define_own_property(
                    Pk, {
                        'value': mappedValue,
                        'writable': True,
                        'enumerable': True,
                        'configurable': True
                    }, False)
            k += 1
        return A

    def filter(this, args):
        array = to_object(this, args.space)
        callbackfn = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if not is_callable(callbackfn):
            raise MakeError('TypeError', 'callbackfn must be a function')
        _this = get_arg(args, 1)
        k = 0
        res = []
        while k < arr_len:
            if array.has_property(unicode(k)):
                kValue = array.get(unicode(k))
                if to_boolean(
                        callbackfn.call(_this, (kValue, float(k), array))):
                    res.append(kValue)
            k += 1
        return args.space.ConstructArray(res)

    def reduce(this, args):
        array = to_object(this, args.space)
        callbackfn = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if not is_callable(callbackfn):
            raise MakeError('TypeError', 'callbackfn must be a function')
        if not arr_len and len(args) < 2:
            raise MakeError('TypeError',
                            'Reduce of empty array with no initial value')
        k = 0
        accumulator = undefined
        if len(args) > 1:  # initial value present
            accumulator = args[1]
        else:
            kPresent = False
            while not kPresent and k < arr_len:
                kPresent = array.has_property(unicode(k))
                if kPresent:
                    accumulator = array.get(unicode(k))
                k += 1
            if not kPresent:
                raise MakeError('TypeError',
                                'Reduce of empty array with no initial value')
        while k < arr_len:
            if array.has_property(unicode(k)):
                kValue = array.get(unicode(k))
                accumulator = callbackfn.call(
                    undefined, (accumulator, kValue, float(k), array))
            k += 1
        return accumulator

    def reduceRight(this, args):
        array = to_object(this, args.space)
        callbackfn = get_arg(args, 0)
        arr_len = js_arr_length(array)
        if not is_callable(callbackfn):
            raise MakeError('TypeError', 'callbackfn must be a function')
        if not arr_len and len(args) < 2:
            raise MakeError('TypeError',
                            'Reduce of empty array with no initial value')
        k = arr_len - 1
        accumulator = undefined

        if len(args) > 1:  # initial value present
            accumulator = args[1]
        else:
            kPresent = False
            while not kPresent and k >= 0:
                kPresent = array.has_property(unicode(k))
                if kPresent:
                    accumulator = array.get(unicode(k))
                k -= 1
            if not kPresent:
                raise MakeError('TypeError',
                                'Reduce of empty array with no initial value')
        while k >= 0:
            if array.has_property(unicode(k)):
                kValue = array.get(unicode(k))
                accumulator = callbackfn.call(
                    undefined, (accumulator, kValue, float(k), array))
            k -= 1
        return accumulator


def sort_compare(a, b, comp):
    if a is None:
        if b is None:
            return 0
        return 1
    if b is None:
        if a is None:
            return 0
        return -1
    if is_undefined(a):
        if is_undefined(b):
            return 0
        return 1
    if is_undefined(b):
        if is_undefined(a):
            return 0
        return -1
    if comp is not None:
        res = comp.call(undefined, (a, b))
        return to_int(res)
    x, y = to_string(a), to_string(b)
    if x < y:
        return -1
    elif x > y:
        return 1
    return 0



INITIALISED = False
babel = None
babelPresetEs2015 = None


def js6_to_js5(code):
    global INITIALISED, babel, babelPresetEs2015
    if not INITIALISED:
        import signal, warnings, time
        warnings.warn(
            '\nImporting babel.py for the first time - this can take some time. \nPlease note that currently Javascript 6 in Js2Py is unstable and slow. Use only for tiny scripts!'
        )

        from .babel import babel as _babel
        babel = _babel.Object.babel
        babelPresetEs2015 = _babel.Object.babelPresetEs2015

        # very weird hack. Somehow this helps babel to initialise properly!
        try:
            babel.transform('warmup', {'presets': {}})
            signal.alarm(2)

            def kill_it(a, b):
                raise KeyboardInterrupt('Better work next time!')

            signal.signal(signal.SIGALRM, kill_it)
            babel.transform('stuckInALoop', {
                'presets': babelPresetEs2015
            }).code
            for n in range(3):
                time.sleep(1)
        except:
            print("Initialised babel!")
        INITIALISED = True
    return babel.transform(code, {'presets': babelPresetEs2015}).code


if __name__ == '__main__':
    print(js6_to_js5('obj={}; obj.x = function() {return () => this}'))
    print()
    print(js6_to_js5('const a = 1;'))



from ..base import *
# python 3 support
import six
if six.PY3:
    unichr = chr


@Js
def fromCharCode():
    args = arguments.to_list()
    res = u''
    for e in args:
        res += unichr(e.to_uint16())
    return this.Js(res)


fromCharCode.own['length']['value'] = Js(1)

String.define_own_property(
    'fromCharCode', {
        'value': fromCharCode,
        'enumerable': False,
        'writable': True,
        'configurable': True
    })

String.define_own_property(
    'prototype', {
        'value': StringPrototype,
        'enumerable': False,
        'writable': False,
        'configurable': False
    })

StringPrototype.define_own_property('constructor', {
    'value': String,
    'enumerable': False,
    'writable': True,
    'configurable': True
})



from __future__ import unicode_literals
from ..conversions import *
from ..func_utils import *


class ErrorPrototype:
    def toString(this, args):
        if Type(this) != 'Object':
            raise MakeError('TypeError',
                            'Error.prototype.toString called on non-object')
        name = this.get('name')
        name = u'Error' if is_undefined(name) else to_string(name)
        msg = this.get('message')
        msg = '' if is_undefined(msg) else to_string(msg)
        return name + (name and msg and ': ') + msg



from __future__ import unicode_literals

from ..conversions import *
from ..func_utils import *


class BooleanPrototype:
    def toString(this, args):
        if GetClass(this) != 'Boolean':
            raise MakeError('TypeError',
                            'Boolean.prototype.toString is not generic')
        if is_object(this):
            this = this.value
        return u'true' if this else u'false'

    def valueOf(this, args):
        if GetClass(this) != 'Boolean':
            raise MakeError('TypeError',
                            'Boolean.prototype.valueOf is not generic')
        if is_object(this):
            this = this.value
        return this



# this is based on jsarray.py

import six
try:
    import numpy
except:
    pass

if six.PY3:
    xrange = range
    import functools


def to_arr(this):
    """Returns Python array from Js array"""
    return [this.get(str(e)) for e in xrange(len(this))]


ARR_STACK = set({})


class TypedArrayPrototype:
    def toString():
        # this function is wrong
        func = this.get('join')
        if not func.is_callable():

            @this.Js
            def func():
                return '[object %s]' % this.Class

        return func.call(this, ())

    def toLocaleString(locales=None, options=None):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        # separator is simply a comma ','
        if not arr_len:
            return ''
        res = []
        for i in xrange(arr_len):
            element = array[str(i)]
            if element.is_undefined() or element.is_null():
                res.append('')
            else:
                cand = element.to_object()
                str_func = element.get('toLocaleString')
                if not str_func.is_callable():
                    raise this.MakeError(
                        'TypeError',
                        'toLocaleString method of item at index %d is not callable'
                        % i)
                res.append(element.callprop('toLocaleString').value)
        return ','.join(res)

    def join(separator):
        ARR_STACK.add(this)
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        separator = ',' if separator.is_undefined() else separator.to_string(
        ).value
        elems = []
        for e in xrange(arr_len):
            elem = array.get(str(e))
            if elem in ARR_STACK:
                s = ''
            else:
                s = elem.to_string().value
            elems.append(
                s if not (elem.is_undefined() or elem.is_null()) else '')
        res = separator.join(elems)
        ARR_STACK.remove(this)
        return res

    def reverse():
        array = this.to_object()  # my own algorithm
        vals = to_arr(array)
        has_props = [array.has_property(str(e)) for e in xrange(len(array))]
        vals.reverse()
        has_props.reverse()
        for i, val in enumerate(vals):
            if has_props[i]:
                array.put(str(i), val)
            else:
                array.delete(str(i))
        return array

    def slice(start, end):  # todo check
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        relative_start = start.to_int()
        k = max((arr_len + relative_start), 0) if relative_start < 0 else min(
            relative_start, arr_len)
        relative_end = arr_len if end.is_undefined() else end.to_int()
        final = max((arr_len + relative_end), 0) if relative_end < 0 else min(
            relative_end, arr_len)
        res = []
        n = 0
        while k < final:
            pk = str(k)
            if array.has_property(pk):
                res.append(array.get(pk))
            k += 1
            n += 1
        return res

    def sort(cmpfn):
        if not this.Class in ('Array', 'Arguments'):
            return this.to_object()  # do nothing
        arr = []
        for i in xrange(len(this)):
            arr.append(this.get(six.text_type(i)))

        if not arr:
            return this
        if not cmpfn.is_callable():
            cmpfn = None
        cmp = lambda a, b: sort_compare(a, b, cmpfn)
        if six.PY3:
            key = functools.cmp_to_key(cmp)
            arr.sort(key=key)
        else:
            arr.sort(cmp=cmp)
        for i in xrange(len(arr)):
            this.put(six.text_type(i), arr[i])

        return this

    def indexOf(searchElement):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if arr_len == 0:
            return -1
        if len(arguments) > 1:
            n = arguments[1].to_int()
        else:
            n = 0
        if n >= arr_len:
            return -1
        if n >= 0:
            k = n
        else:
            k = arr_len - abs(n)
            if k < 0:
                k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                elementK = array.get(str(k))
                if searchElement.strict_equality_comparison(elementK):
                    return k
            k += 1
        return -1

    def lastIndexOf(searchElement):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if arr_len == 0:
            return -1
        if len(arguments) > 1:
            n = arguments[1].to_int()
        else:
            n = arr_len - 1
        if n >= 0:
            k = min(n, arr_len - 1)
        else:
            k = arr_len - abs(n)
        while k >= 0:
            if array.has_property(str(k)):
                elementK = array.get(str(k))
                if searchElement.strict_equality_comparison(elementK):
                    return k
            k -= 1
        return -1

    def every(callbackfn):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                if not callbackfn.call(
                        T, (kValue, this.Js(k), array)).to_boolean().value:
                    return False
            k += 1
        return True

    def some(callbackfn):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                if callbackfn.call(
                        T, (kValue, this.Js(k), array)).to_boolean().value:
                    return True
            k += 1
        return False

    def forEach(callbackfn):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                callbackfn.call(T, (kValue, this.Js(k), array))
            k += 1

    def map(callbackfn):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        A = this.Js([])
        k = 0
        while k < arr_len:
            Pk = str(k)
            if array.has_property(Pk):
                kValue = array.get(Pk)
                mappedValue = callbackfn.call(T, (kValue, this.Js(k), array))
                A.define_own_property(
                    Pk, {
                        'value': mappedValue,
                        'writable': True,
                        'enumerable': True,
                        'configurable': True
                    })
            k += 1
        return A

    def filter(callbackfn):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        T = arguments[1]
        res = []
        k = 0
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                if callbackfn.call(
                        T, (kValue, this.Js(k), array)).to_boolean().value:
                    res.append(kValue)
            k += 1
        return res  # converted to js array automatically

    def reduce(callbackfn):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        if not arr_len and len(arguments) < 2:
            raise this.MakeError(
                'TypeError', 'Reduce of empty array with no initial value')
        k = 0
        if len(arguments) > 1:  # initial value present
            accumulator = arguments[1]
        else:
            kPresent = False
            while not kPresent and k < arr_len:
                kPresent = array.has_property(str(k))
                if kPresent:
                    accumulator = array.get(str(k))
                k += 1
            if not kPresent:
                raise this.MakeError(
                    'TypeError', 'Reduce of empty array with no initial value')
        while k < arr_len:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                accumulator = callbackfn.call(
                    this.undefined, (accumulator, kValue, this.Js(k), array))
            k += 1
        return accumulator

    def reduceRight(callbackfn):
        array = this.to_object()
        arr_len = array.get("length").to_uint32()
        if not callbackfn.is_callable():
            raise this.MakeError('TypeError', 'callbackfn must be a function')
        if not arr_len and len(arguments) < 2:
            raise this.MakeError(
                'TypeError', 'Reduce of empty array with no initial value')
        k = arr_len - 1
        if len(arguments) > 1:  # initial value present
            accumulator = arguments[1]
        else:
            kPresent = False
            while not kPresent and k >= 0:
                kPresent = array.has_property(str(k))
                if kPresent:
                    accumulator = array.get(str(k))
                k -= 1
            if not kPresent:
                raise this.MakeError(
                    'TypeError', 'Reduce of empty array with no initial value')
        while k >= 0:
            if array.has_property(str(k)):
                kValue = array.get(str(k))
                accumulator = callbackfn.call(
                    this.undefined, (accumulator, kValue, this.Js(k), array))
            k -= 1
        return accumulator


def sort_compare(a, b, comp):
    if a is None:
        if b is None:
            return 0
        return 1
    if b is None:
        if a is None:
            return 0
        return -1
    if a.is_undefined():
        if b.is_undefined():
            return 0
        return 1
    if b.is_undefined():
        if a.is_undefined():
            return 0
        return -1
    if comp is not None:
        res = comp.call(a.undefined, (a, b))
        return res.to_int()
    x, y = a.to_string(), b.to_string()
    if x < y:
        return -1
    elif x > y:
        return 1
    return 0



# coding=utf-8
from .translators import translate_js, DEFAULT_HEADER
from .es6 import js6_to_js5
import sys
import time
import json
import six
import os
import hashlib
import codecs

__all__ = [
    'EvalJs', 'translate_js', 'import_js', 'eval_js', 'translate_file',
    'eval_js6', 'translate_js6', 'run_file', 'disable_pyimport',
    'get_file_contents', 'write_file_contents'
]
DEBUG = False


def disable_pyimport():
    import pyjsparser.parser
    pyjsparser.parser.ENABLE_PYIMPORT = False


def path_as_local(path):
    if os.path.isabs(path):
        return path
    # relative to cwd
    return os.path.join(os.getcwd(), path)


def import_js(path, lib_name, globals):
    """Imports from javascript source file.
      globals is your globals()"""
    with codecs.open(path_as_local(path), "r", "utf-8") as f:
        js = f.read()
    e = EvalJs()
    e.execute(js)
    var = e.context['var']
    globals[lib_name] = var.to_python()


def get_file_contents(path_or_file):
    if hasattr(path_or_file, 'read'):
        js = path_or_file.read()
    else:
        with codecs.open(path_as_local(path_or_file), "r", "utf-8") as f:
            js = f.read()
    return js


def write_file_contents(path_or_file, contents):
    if hasattr(path_or_file, 'write'):
        path_or_file.write(contents)
    else:
        with codecs.open(path_as_local(path_or_file), "w", "utf-8") as f:
            f.write(contents)


def translate_file(input_path, output_path):
    '''
    Translates input JS file to python and saves the it to the output path.
    It appends some convenience code at the end so that it is easy to import JS objects.

    For example we have a file 'example.js' with:   var a = function(x) {return x}
    translate_file('example.js', 'example.py')

    Now example.py can be easily importend and used:
    >>> from example import example
    >>> example.a(30)
    30
    '''
    js = get_file_contents(input_path)

    py_code = translate_js(js)
    lib_name = os.path.basename(output_path).split('.')[0]
    head = '__all__ = [%s]\n\n# Don\'t look below, you will not understand this Python code :) I don\'t.\n\n' % repr(
        lib_name)
    tail = '\n\n# Add lib to the module scope\n%s = var.to_python()' % lib_name
    out = head + py_code + tail
    write_file_contents(output_path, out)


def run_file(path_or_file, context=None):
    ''' Context must be EvalJS object. Runs given path as a JS program. Returns (eval_value, context).
    '''
    if context is None:
        context = EvalJs()
    if not isinstance(context, EvalJs):
        raise TypeError('context must be the instance of EvalJs')
    eval_value = context.eval(get_file_contents(path_or_file))
    return eval_value, context


def eval_js(js):
    """Just like javascript eval. Translates javascript to python,
       executes and returns python object.
       js is javascript source code

       EXAMPLE:
        >>> import js2py
        >>> add = js2py.eval_js('function add(a, b) {return a + b}')
        >>> add(1, 2) + 3
        6
        >>> add('1', 2, 3)
        u'12'
        >>> add.constructor
        function Function() { [python code] }

       NOTE: For Js Number, String, Boolean and other base types returns appropriate python BUILTIN type.
       For Js functions and objects, returns Python wrapper - basically behaves like normal python object.
       If you really want to convert object to python dict you can use to_dict method.
       """
    e = EvalJs()
    return e.eval(js)


def eval_js6(js):
    """Just like eval_js but with experimental support for js6 via babel."""
    return eval_js(js6_to_js5(js))


def translate_js6(js):
    """Just like translate_js but with experimental support for js6 via babel."""
    return translate_js(js6_to_js5(js))


class EvalJs(object):
    """This class supports continuous execution of javascript under same context.

        >>> ctx = EvalJs()
        >>> ctx.execute('var a = 10;function f(x) {return x*x};')
        >>> ctx.f(9)
        81
        >>> ctx.a
        10

        context is a python dict or object that contains python variables that should be available to JavaScript
        For example:
        >>> ctx = EvalJs({'a': 30})
        >>> ctx.execute('var x = a')
        >>> ctx.x
        30

        You can enable JS require function via enable_require. With this feature enabled you can use js modules
        from npm, for example:
        >>> ctx = EvalJs(enable_require=True)
        >>> ctx.execute("var esprima = require('esprima');")
        >>> ctx.execute("esprima.parse('var a = 1')")

       You can run interactive javascript console with console method!"""

    def __init__(self, context={}, enable_require=False):
        self.__dict__['_context'] = {}
        exec (DEFAULT_HEADER, self._context)
        self.__dict__['_var'] = self._context['var'].to_python()

        if enable_require:
            def _js_require_impl(npm_module_name):
                from .node_import import require
                from .base import to_python
                return require(to_python(npm_module_name), context=self._context)
            setattr(self._var, 'require', _js_require_impl)

        if not isinstance(context, dict):
            try:
                context = context.__dict__
            except:
                raise TypeError(
                    'context has to be either a dict or have __dict__ attr')
        for k, v in six.iteritems(context):
            setattr(self._var, k, v)

    def execute(self, js=None, use_compilation_plan=False):
        """executes javascript js in current context

        During initial execute() the converted js is cached for re-use. That means next time you
        run the same javascript snippet you save many instructions needed to parse and convert the
        js code to python code.

        This cache causes minor overhead (a cache dicts is updated) but the Js=>Py conversion process
        is typically expensive compared to actually running the generated python code.

        Note that the cache is just a dict, it has no expiration or cleanup so when running this
        in automated situations with vast amounts of snippets it might increase memory usage.
        """
        try:
            cache = self.__dict__['cache']
        except KeyError:
            cache = self.__dict__['cache'] = {}
        hashkey = hashlib.md5(js.encode('utf-8')).digest()
        try:
            compiled = cache[hashkey]
        except KeyError:
            code = translate_js(
                js, '', use_compilation_plan=use_compilation_plan)
            compiled = cache[hashkey] = compile(code, '<EvalJS snippet>',
                                                'exec')
        exec (compiled, self._context)

    def eval(self, expression, use_compilation_plan=False):
        """evaluates expression in current context and returns its value"""
        code = 'PyJsEvalResult = eval(%s)' % json.dumps(expression)
        self.execute(code, use_compilation_plan=use_compilation_plan)
        return self['PyJsEvalResult']

    def execute_debug(self, js):
        """executes javascript js in current context
        as opposed to the (faster) self.execute method, you can use your regular debugger
        to set breakpoints and inspect the generated python code
        """
        code = translate_js(js, '')
        # make sure you have a temp folder:
        filename = 'temp' + os.sep + '_' + hashlib.md5(
            code.encode("utf-8")).hexdigest() + '.py'
        try:
            with open(filename, mode='w') as f:
                f.write(code)
            with open(filename, "r") as f:
                pyCode = compile(f.read(), filename, 'exec')
                exec(pyCode, self._context)
                
        except Exception as err:
            raise err
        finally:
            os.remove(filename)
            try:
                os.remove(filename + 'c')
            except:
                pass

    def eval_debug(self, expression):
        """evaluates expression in current context and returns its value
        as opposed to the (faster) self.execute method, you can use your regular debugger
        to set breakpoints and inspect the generated python code
        """
        code = 'PyJsEvalResult = eval(%s)' % json.dumps(expression)
        self.execute_debug(code)
        return self['PyJsEvalResult']

    @property
    def context(self):
        return self._context
    
    def __getattr__(self, var):
        return getattr(self._var, var)

    def __getitem__(self, var):
        return getattr(self._var, var)

    def __setattr__(self, var, val):
        return setattr(self._var, var, val)

    def __setitem__(self, var, val):
        return setattr(self._var, var, val)

    def console(self):
        """starts to interact (starts interactive console) Something like code.InteractiveConsole"""
        while True:
            if six.PY2:
                code = raw_input('>>> ')
            else:
                code = input('>>>')
            try:
                print(self.eval(code))
            except KeyboardInterrupt:
                break
            except Exception as e:
                import traceback
                if DEBUG:
                    sys.stderr.write(traceback.format_exc())
                else:
                    sys.stderr.write('EXCEPTION: ' + str(e) + '\n')
                time.sleep(0.01)



# python 3 support
import six
if six.PY3:
    basestring = str
    long = int
    xrange = range
    unicode = str


class FunctionPrototype:
    def toString():
        if not this.is_callable():
            raise TypeError('toString is not generic!')
        args = ', '.join(this.code.__code__.co_varnames[:this.argcount])
        return 'function %s(%s) ' % (this.func_name, args) + this.source

    def call():
        arguments_ = arguments
        if not len(arguments):
            obj = this.Js(None)
        else:
            obj = arguments[0]
        if len(arguments) <= 1:
            args = ()
        else:
            args = tuple([arguments_[e] for e in xrange(1, len(arguments_))])
        return this.call(obj, args)

    def apply():
        if not len(arguments):
            obj = this.Js(None)
        else:
            obj = arguments[0]
        if len(arguments) <= 1:
            args = ()
        else:
            appl = arguments[1]
            args = tuple([appl[e] for e in xrange(len(appl))])
        return this.call(obj, args)

    def bind(thisArg):
        arguments_ = arguments
        target = this
        if not target.is_callable():
            raise this.MakeError(
                'Object must be callable in order to be used with bind method')
        if len(arguments) <= 1:
            args = ()
        else:
            args = tuple([arguments_[e] for e in xrange(1, len(arguments_))])
        return this.PyJsBoundFunction(target, thisArg, args)
